// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#version 330

#define PI 3.14159265359

#define og_oneOverTwoPi (1.0/(2.0*PI))
#define og_oneOverPi (1.0/PI)

uniform vec3 eyePosition;
uniform mat4 model;
uniform mat4 projection;
uniform mat4 view;

in vec4 worldPosition;
out vec4 color;

struct Intersection
{
    bool  Intersects;
    float NearTime;         // Along ray
    float FarTime;          // Along ray
};

//
// Assumes ellipsoid is at (0, 0, 0)
//
Intersection RayIntersectEllipsoid(vec3 rayOrigin, vec3 rayOriginSquared, vec3 rayDirection, vec3 oneOverEllipsoidRadiiSquared)
{
    float a = dot(rayDirection * rayDirection, oneOverEllipsoidRadiiSquared);
    float b = 2.0 * dot(rayOrigin * rayDirection, oneOverEllipsoidRadiiSquared);
    float c = dot(rayOriginSquared, oneOverEllipsoidRadiiSquared) - 1.0;
    float discriminant = b * b - 4.0 * a * c;

    if (discriminant < 0.0)
    {
        return Intersection(false, 0.0, 0.0);
    }
    else if (discriminant == 0.0)
    {
        float time = -0.5 * b / a;
        return Intersection(true, time, time);
    }

    float t = -0.5 * (b + (b > 0.0 ? 1.0 : -1.0) * sqrt(discriminant));
    float root1 = t / a;
    float root2 = c / t;

    return Intersection(true, min(root1, root2), max(root1, root2));
}

float ComputeWorldPositionDepth(vec3 position, mat4 modelZToClipCoordinates)
{ 
    vec4 v = modelZToClipCoordinates * vec4(position, 1);   // clip coordinates
    v.z /= v.w;                                             // normalized device coordinates
    v.z = (v.z + 1.0) * 0.5;
    return v.z;
}

vec2 ComputeTextureCoordinates(vec3 normal)
{
    return vec2(-(atan(normal.x, normal.y) / PI + 0.5) * 0.5, -(asin(normal.z) / PI + 0.5));
}

void main()
{
    vec3 eyeOrigin = eyePosition;
    vec3 rayDirection = normalize(worldPosition.xyz - eyeOrigin);

    Intersection i = RayIntersectEllipsoid(eyeOrigin, vec3(eyeOrigin.x*eyeOrigin.x,eyeOrigin.y*eyeOrigin.y,eyeOrigin.z*eyeOrigin.z), rayDirection, vec3(1.0));

    if (i.Intersects)
    {
        vec3 position = normalize(eyeOrigin + rayDirection * i.NearTime);
        //color = vec4(0.2);
        color = vec4(0.0);
        position = eyeOrigin + (mix(i.NearTime, i.FarTime, 0.5) * rayDirection);
        gl_FragDepth = ComputeWorldPositionDepth(position, projection * view);
    }
    else
    {
        discard;
    }
}
