// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#version 330 core

layout(location = 0) in vec3 position;
layout(location = 1) in vec3 neighbour;
// Position of the vertex in the current streamline [0, streamlineLength-1]
layout(location = 2) in float localID;

uniform sampler2D texture0; // physical
uniform sampler2D texture1; // mask
uniform sampler2D texture2; // gcp up vector
uniform sampler2D texture3; // gcp Aspect Ratio

uniform vec2 uvMin;
uniform vec2 uvRange;
uniform float speed;
uniform float headID;
uniform vec4 scaleFactor;
uniform vec4 offset;
uniform float glFactor;
uniform vec2 uvTexelStep;
uniform float rangeMax;
uniform bvec2 seamless;
uniform float ttl; // sec
uniform float fade; // sec 

// New position to pass to the fragment shader
out vec3 position0;

void main()
{
	// Check if it's the streamline head or a part of the body
    if (localID == headID)
    {
        // It's the head, compute her new position with the value of the dataset.
		vec2 pos = position.xy;

		vec2 particlePos = (pos - uvMin) / uvRange;

        //for (int i = 0; i < 20; ++i)
        for (int i = 0; i < 1; ++i)
        {
		    // Read the mask
		    float mask = texture(texture1, position.xy).x;

		    // Check if there are data here.
		    if (mask > 0.0)
		    {
			    // Data is available
			    // Read the vector in the dataset
			    vec2 dir = ((texture(texture0, particlePos).rg * glFactor) * scaleFactor.rg) + offset.rg;

			    float value = 1.0;
                if (0.0 < rangeMax)
                {
                    value = min(length(dir), rangeMax) / rangeMax;
                }

                if (0.0 >= value)
                {
                    pos = position.xy;
                    break;
                }

			    // North and East expressed in texels
                vec4 toto = texture(texture2, position.xy);
			    vec2 vY = toto.xy;  // North
			    vec2 vX = toto.zw;  // East

			    dir = normalize(dir.x * vX + dir.y * vY) * value;

                vec2 aspectRatio = texture(texture3, position.xy).rg;

                // change length depending of the latitude
                dir = dir * aspectRatio.y;

                // apply local ratio of the gcp quad
                dir.y = dir.y * aspectRatio.x;
                
			    // Compute the new position
			    pos = pos + dir * uvTexelStep * speed;

			    if (seamless.x)
                {
                    if (pos.x >= 1.0)
                    {
                        pos.x -= 1.0;
                    }
                    else if (pos.x < 0.0)
                    {
                        pos.x += 1.0;
                    }
                }
                else if (pos.x >= 1.0 || pos.x <= 0.0)
                {
                    pos = position.xy;
                    break;
                }

                if (seamless.y)
                {
                    if (pos.y >= 1.0)
                    {
                        pos.y -= 1.0;
                    }
                    else if (pos.y < 0.0)
                    {
                        pos.y += 1.0;
                    }
                }
                else if (pos.y >= 1.0 || pos.y <= 0.0)
                {
                    pos = position.xy;
                    break;
                }

                pos.x = min(pos.x, 1.0);
                pos.y = min(pos.y, 1.0);
		    }
		    else
		    {
			    particlePos = pos;
                break;
		    }
        }

		// Save the result
		position0.xy = pos;
	}
	else
    {
        //position0 = neighbour;
        vec2 pos = position.xy;
        position0.xy = pos.xy + ((neighbour.xy - pos.xy));
    }

    //position0.z = position0.z + 0.1;
    position0.z = position.z;
}
