// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#version 330 core

layout(location = 0) in vec3 position;
layout(location = 1) in vec3 previous;
layout(location = 2) in vec3 next;
layout(location = 3) in vec2 direction; 
layout(location = 4) in float data; 

uniform mat4    mvp;
uniform float   thickness;
uniform vec2    viewport;
uniform vec2    range;

out vec2 texCoord0;
flat out float segmentLength;
out float lineWidth;
out float data0;

vec2 project(vec4 P)
{
    vec2 p = 0.5 + (P.xyz/P.w).xy * 0.5;
    return p * viewport;
}

// Project from the screen space to the world space
vec4 unproject(vec2 p, float z, float w)
{
    vec4 P = vec4( w*((p/viewport)*2.0 - 1.0), z, w);
    return P;
}

void main() 
{
    mat4 projViewModel = mvp;

    // vertex 
    vec4 currentProjected    = projViewModel * vec4(position, 1.0);

    //get 2D screen space with W divide and aspect correction
    vec2 currentScreen    = project(currentProjected);
    vec2 previousScreen   = project(projViewModel * vec4(previous, 1.0));
    vec2 nextScreen       = project(projViewModel * vec4(next, 1.0));

    lineWidth = thickness / 2.0;
    vec2 orientation = vec2(direction.x, direction.y > 0 ? 1.0 : -1.0);

    //vec2 normal;

    int type = int(abs(direction.y));

    //starting point uses (next - current)
    vec2 dir = vec2(0.0);
    //vec3 dir3 = vec3(0.0);

    vec2 X = vec2(1.0,0.0);
    vec2 Y = vec2(0.0,1.0);

    if (type == 1) 
    {
        dir = normalize(nextScreen - currentScreen);
        segmentLength = distance(currentScreen, nextScreen);
    } 
    //ending point uses (current - previous)
    else if (type == 3) 
    {
        dir = normalize(currentScreen - previousScreen);
        segmentLength = distance(currentScreen, previousScreen);
    }
    //somewhere in middle, needs a join
    else 
    {
        // Compute segment direction
        vec2 dirA = normalize(currentScreen - previousScreen);
        vec2 dirB = normalize(nextScreen - currentScreen);

        if (direction.x < 0.0)
        {
            dir = dirB;
            segmentLength = distance(currentScreen, nextScreen);
        }
        else
        {
            dir = dirA;
            segmentLength = distance(currentScreen, previousScreen);
        }
    }

    X = dir * orientation.x;
    Y = vec2(-dir.y, dir.x) * orientation.y;

    vec2 normal = (X + Y);

    float antialias = 1.0;  
    float decal = (1.25) * antialias;
    normal *= lineWidth + decal;

    texCoord0.x = (lineWidth + decal) * orientation.x + (max(0, orientation.x) * segmentLength);
    texCoord0.y = (lineWidth + decal) * orientation.y;

    data0 = (clamp(data, range.x, range.y) - range.x) / (range.y - range.x);

    gl_Position = unproject(currentScreen + normal, currentProjected.z, currentProjected.w);
}

/*void main() 
{
    vec2 halfViewport = viewport / 2.0;
    mat4 projViewModel = mvp;

    vec4 previousProjected    = projViewModel * vec4(previous, 1.0);
    vec4 currentProjected     = projViewModel * vec4(position, 1.0);
    vec4 nextProjected        = projViewModel * vec4(next, 1.0);

    //get 2D screen space with W divide and aspect correction
    vec2 currentScreen    = currentProjected.xy / currentProjected.w;
    vec2 previousScreen   = previousProjected.xy / previousProjected.w;
    vec2 nextScreen       = nextProjected.xy / nextProjected.w;

    float len = thickness / halfViewport.x;
    vec2 orientation = vec2(direction.x, direction.y > 0 ? 1.0 : -1.0);

    //vec2 normal;

    int type = int(abs(direction.y));

    //starting point uses (next - current)
    vec2 dir = vec2(0.0);
    //vec3 dir3 = vec3(0.0);

    vec2 X = vec2(1.0,0.0);
    vec2 Y = vec2(0.0,1.0);

    if (type == 1) 
    {
      color0 = vec3(1,0,0);
      dir = normalize(nextScreen - currentScreen);
      segmentLength = distance(currentScreen, nextScreen);

      //dir3 = normalize(next - position);
    } 
    //ending point uses (current - previous)
    else if (type == 3) 
    {
      color0 = vec3(0,1,0);

      dir = normalize(currentScreen - previousScreen);
      segmentLength = distance(currentScreen, previousScreen);

      //dir3 = normalize(position - previous);
    }
    //somewhere in middle, needs a join
    else 
    {
      color0 = vec3(0,0,1);

      // Compute segment direction
      vec2 dirA = normalize(currentScreen - previousScreen);
      vec2 dirB = normalize(nextScreen - currentScreen);

      //vec3 dirB3 = normalize(next - position);
      //vec3 dirA3 = normalize(position - previous);


    
      if (miter == 0)
      {
          vec2 dirB = normalize(nextScreen - currentScreen);

          vec2 tangent = normalize(dirA + dirB);
          vec2 perp = vec2(-dirA.y, dirA.x);
          vec2 miter = vec2(-tangent.y, tangent.x);
          dir = tangent;
          len = (thickness / 800) / dot(miter, perp);
          //len = (thickness) / dot(miter, perp);
      }
      else
      {
          if (direction.x < 0.0)
          {
              dir = dirB;
              segmentLength = distance(currentScreen, nextScreen);

              //dir3 = dirB3;
          }
          else
          {
              dir = dirA;
              segmentLength = distance(currentScreen, previousScreen);

              //dir3 = dirA3;
          }
      }
    }

    //vec3 normal3 = cross(dir3.xyz, normalize(position));
    //vec3 X3 = dir3 * orientation.x;
    //vec3 Y3 = normal3 * orientation.y;

    //vec3 offset3 = X3 + Y3;

    X = dir * orientation.x;
    Y = vec2(-dir.y, dir.x) * orientation.y;

    vec2 normal = (X + Y);

    float antialias = 1.0;  
    float decal = (1.25 / halfViewport.x) * antialias;
    normal *= len/2.0 + decal;

    texCoord0.x = (len/2.0  + decal) * orientation.x + (max(0, orientation.x) * segmentLength);
    texCoord0.y = (len/2.0  + decal) * orientation.y;

    gl_Position = vec4(currentScreen + normal, 0.0, 1.0);
    //gl_Position = vec4((currentScreen + normal) * currentProjected.w, currentProjected.z, currentProjected.w);
    //gl_Position = vec4(currentScreen + normal, currentProjected.z / currentProjected.w, 1.0);
    //gl_Position = projViewModel * vec4(position + offset3 * 0.001, 1.0);
}
*/
