// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2025 OceanDataLab
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
#version 330 core

precision highp float;

layout(location = 0) in vec2 position;
layout(location = 2) in vec2 direction;

uniform sampler2D texture0; // mask
uniform sampler2D texture3; // physical data
uniform sampler2D texture5; // gcp 

uniform mat4 projViewModel;
uniform vec2 viewport;
// half diagonal size, square edge = (arrowTileSize * 2) / sqrt(2)
uniform float arrowTileSize;
uniform mat4 invProjViewModel;
uniform vec4 scaleFactor;
uniform vec4 offset;
uniform float glFactor;
uniform vec2 uvMin;
uniform vec2 uvRange;
uniform vec2 offsetUV;
uniform float dist;

out vec2 texCoord0;
flat out float angle;
flat out float windSpeedKnot;

vec2 project(vec4 P)
{
    vec2 p = 0.5 + (P.xyz/P.w).xy * 0.5;
    return p * viewport;
}

// Project from the screen space to the world space
vec4 unproject(vec2 p, float z, float w)
{
    vec4 P = vec4( w*((p/viewport)*2.0 - 1.0), z, w);
    return P;
}

void main()
{
    vec2 posOffset = position + offsetUV;

    vec3 particlePos3 = normalize(texture(texture5, posOffset).xyz);

    vec4 currentProjected = projViewModel * vec4(particlePos3, 1.0);
    vec2 currentScreen    = project(currentProjected);

    // Compute pixel coordinates in Normalized Device Coordinates space [0, 1].
    vec2 pixelCoord = 0.5 + (currentProjected.xyz/currentProjected.w).xy * 0.5;

    // Check screen limits, the vertex can be outside of the screen, but there are no data to read.
    if (pixelCoord.x < 0.0 || pixelCoord.y < 0.0 || pixelCoord.x > 1.0 || pixelCoord.y > 1.0)
    {
        // Outside vertex, do nothing.
        gl_Position = currentProjected;
        return;
    }

    float mask = texture(texture0, posOffset).r;

    if (mask < 1.0)
    {
        gl_Position = vec4(-2.0, 0.0, 0.0, 1.0);
        return;
    }

    vec2 particleDataUV = (posOffset - uvMin) / uvRange;

    // Read physical data and uncompress it
    vec2 uv = ((texture(texture3, particleDataUV).rg * glFactor) * scaleFactor.rg) + offset.rg;
        
    windSpeedKnot = length(uv.xy); 

    // Compute the position of the current corner vertex in screen space
    vec2 dirNorm = normalize(direction);

    // Compute the texture coordinates.
    texCoord0 = clamp(direction, 0.0, 1.0);

    // Create local axis system
    vec3 yaxis = vec3(0.0, 0.0, 1.0); // centre de la terre vers le pole nord
    vec3 zaxis = normalize(particlePos3); // particule pos
    vec3 xaxis = normalize(cross(zaxis, yaxis));
    yaxis = normalize(cross(zaxis, xaxis));

    // Compute UV axis in world space
    vec3 xUVaxis = normalize(xaxis * uv.x + yaxis * uv.y);
    vec3 yUVaxis = normalize(xaxis * (-uv.y) + yaxis * uv.x);

    // Compute the position of the current vertex in worldspace
    vec3 corner = particlePos3 + (xUVaxis * dirNorm.x + yUVaxis * dirNorm.y) * dist;

    gl_Position = projViewModel * vec4(corner, 1.0);
}
